package SliMP3::Control;

# $Id: Control.pm,v 1.50 2003/02/24 19:55:52 blackketter Exp $

# SliMP3 Server Copyright (C) 2001 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License, 
# version 2.

use strict;

use SliMP3::mas3507d;
use SliMP3::Misc;
use SliMP3::Timers;
use SliMP3::Stream;

# the old streaming protocol had a separate "out of band" set of codes to control
# pausing, stopping, starting a new stream, etc. In the new protocol, this logic is
# controlled entirely by the server, and the simplified control codes are now sent
# along with the stream data.
# 
# This module serves as high level glue, providing the basic stream control functions
# independently of which protocol we're using underneath.

# see protocol docs
#	
my %streamControlCode = (
	'old_start',       pack('C', 1),
	'old_startpaused', pack('C', 3),
	'old_pause',       pack('C', 2),
	'old_resume',      pack('C', 4),
	'old_showusage',   pack('C', 8),
);

$SliMP3::Control::maxVolume = 100;

$SliMP3::Control::maxTreble = 100;
$SliMP3::Control::minTreble = 0;
$SliMP3::Control::maxBass = 100;
$SliMP3::Control::minBass = 0;

# The codes come from the MAS3507D datasheet 

my @trebleCodes =(
    'b2c00',    
    'bb400',    
    'c1800',    
    'c6c00',    
    'cbc00',
    'd0400',
    'd5000',
    'd9800',
    'de000',
    'e2800',    
    'e7e00',    
    'ec000',
    'f0c00',
    'f5c00',
    '0fac00',
    '00000',    # midpoint (50% or no adjustment)
    '05400',
    '0ac00',
    '10400',
    '16000',
    '1c000',
    '22000',
    '28400',
    '2ec00',
    '35400',
    '3c000',
    '42c00',
    '49c00',
    '51800',
    '58400',
    '5f800',
);



my @bassCodes =(
    '9e400',    
    'a2800',    
    'a7400',    
    'ac400',    
    'b1800',    
    'b7400',    
    'bd400',    
    'c3c00',    
    'ca400',    
    'd1800',    
    'd8c00',    
    'e0400',    
    'e8000',    
    'efc00',
    'f7c00',
    '00000',  # midpoint (50% or no adjustment)
    '00800',
    '10000',
    '17c00',
    '1f800',
    '27000',
    '2e400',
    '35800',
    '3c000',
    '42800',
    '48800',
    '4e400',
    '53800',
    '58800',
    '5d400',
    '61800',
);


my @prefactorCodes =(
    'e9400',
    'e6800',
    'e3400',
    'dfc00',
    'dc000',
    'd7800',
    'd25c0',
    'cd000',
    'c6c00',
    'bfc00',
    'b8000',
    'af400',
    'a5800',
    '9a400',
    '8e000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000',
    '80000'    
);


#
# initialize the MAS3507D and tell the client to start a new stream
#
sub play {
	my $client = shift;
	my $paused = shift;
	if (!SliMP3::Client::isSliMP3($client)) {
		return 1;
	}
	# make sure the volume is set correctly.
	# this is done in Power::powerUp also
	&volume($client, SliMP3::Prefs::clientGet($client, "volume"));
	
	SliMP3::Protocol::sendClient($client, '2                 '
		.SliMP3::mas3507d::masWrite('config','00002')
		.SliMP3::mas3507d::masWrite('loadconfig')
		.SliMP3::mas3507d::masWrite('plloffset48','5d9e9')
		.SliMP3::mas3507d::masWrite('plloffset44','cecf5')
		.SliMP3::mas3507d::masWrite('setpll')
	);

	# sleep for .05 seconds
	# this little kludge is to prevent overflowing cs8900's RX buf because the client
	# is still processing the i2c commands we just sent.
	select(undef,undef,undef,.05);

	if ($client->revision < 1.3) {
		if ($paused) {
			# if we're synced with other players, use the startpaused mode to let
			# all the players fill their buffers before we unpause them.
			streamControl($client, 'old_startpaused');
		} else {
			# otherwise, let the client decide when to start
			# I guess we could also just use startpaused for both cases...
			# the old streaming protocol needs to die soon anyway.
			streamControl($client, 'old_start');
		}
	} else {
		SliMP3::Stream::newStream($client, $paused);
	}
	return 1;
}

#
# set the MAS3507D volume
#
sub volume {
	my ($client, $volume) = @_;

	if (!SliMP3::Client::isSliMP3($client)) {
		return 1;
	}

	if ($volume > $SliMP3::Control::maxVolume) { $volume = $SliMP3::Control::maxVolume; }
	if ($volume < 0) { $volume = 0; }
	
	# normalize
	$volume = $volume / $SliMP3::Control::maxVolume;
	
	# volume squared seems to correlate better with the linear scale.  I'm sure there's something
	# optimal, but this is better.
	my $level = sprintf('%X', 0xFFFFF - 0x7FFFF * $volume * $volume);

	# FIXME (sean) there's a retarded bug in the way we make the i2c packets.
	# we use space as a delimeter and so it doesn't handle 0x20. 
	$level =~s/20/21/g;

	SliMP3::Protocol::sendClient($client, '2                 '
		. SliMP3::mas3507d::masWrite('ll', $level)
		. SliMP3::mas3507d::masWrite('rr', $level)
	);
	
}

#
# set the MAS3507D treble in the range of -1 to 1
#

sub treble {
	my ($client, $treble) = @_;
	if ($treble > $SliMP3::Control::maxTreble) { $treble = $SliMP3::Control::maxTreble; }
	if ($treble < $SliMP3::Control::minTreble) { $treble = $SliMP3::Control::minTreble; }

	my $index = int($treble / $SliMP3::Control::maxTreble * (scalar(@trebleCodes)-1) + 0.5);
	
	SliMP3::Protocol::sendClient($client, '2                 '
		.SliMP3::mas3507d::masWrite('treble', $trebleCodes[$index])
	);	
	$::d_control && msg("setting new treble value of $treble (index $index)\n"); 
#	SliMP3::Protocol::sendClient($client, '2                 '
#		.SliMP3::mas3507d::masWrite('prefactor', $prefactorCodes{$treble})
#	);	
}

#
# set the MAS3507D bass in the range of -1 to 1
#

sub bass {
	my ($client, $bass) = @_;
	if ($bass > $SliMP3::Control::maxBass) { $bass = $SliMP3::Control::maxBass; }
	if ($bass < $SliMP3::Control::minBass) { $bass = $SliMP3::Control::minBass; }
	
	my $index = int($bass / $SliMP3::Control::maxTreble * (scalar(@bassCodes)-1) + 0.5);
	
	SliMP3::Protocol::sendClient($client, '2                 '
		.SliMP3::mas3507d::masWrite('bass', $bassCodes[$index])
	);	
	
	$::d_control && msg("setting new bass value of $bass (index $index)\n"); 
#	SliMP3::Protocol::sendClient($client, '2                 '
#		.SliMP3::mas3507d::masWrite('prefactor', $prefactorCodes{$bass})
#	);	
}



# fade the volume up or down
# $fade = amount to fade (positive to fade up, negative to fade down)
# $callback is function reference to be called when the fade is complete
# FYI 8 to 10 seems to be a good fade value
my %fvolume;  # keep temporary fade volume for each client
sub fade_volume {
	my($client, $fade, $callback, $callbackargs) = @_;

	if (!SliMP3::Client::isSliMP3($client)) {
		return 1;
	}
	SliMP3::Timers::killTimers($client, \&fade_volume);
	
	my $vol = SliMP3::Prefs::clientGet($client, "volume");
	if ($vol < 0) {
		# the volume is muted, don't fade.
		$callback && (&$callback(@$callbackargs));
		return;
	}
	
	# on the first pass, set temporary fade volume
	if(!$fvolume{$client} && $fade > 0) {
		# fading up, start volume at 0
		$fvolume{$client} = 0;
	} elsif(!$fvolume{$client}) {
		# fading down, start volume at current volume
		$fvolume{$client} = $vol;
	}

	$fvolume{$client} += $fade; # fade volume

	if($fvolume{$client} <= 0 || $fvolume{$client} >= $vol) {
		# done fading
		$::d_ui && msg("fade_volume done.\n");
		$fvolume{$client} = 0; # reset temporary fade volume 
		$callback && (&$callback(@$callbackargs));
	} else {
		$::d_ui && msg("fade_volume - setting volume to $fvolume{$client}\n");
		&volume($client, $fvolume{$client}); # set volume
		SliMP3::Timers::setTimer($client, Time::HiRes::time()+0.025, \&fade_volume, ($fade, $callback, $callbackargs));
	}
}

# mute or un-mute volume as necessary
# A negative volume indicates that the player is muted and should be restored 
# to the absolute value when un-muted.
sub mute {
	my $client = shift;
	
	if (!SliMP3::Client::isSliMP3($client)) {
		return 1;
	}
	my $vol = SliMP3::Prefs::clientGet($client, "volume");
	$vol *= -1;
			
	if($vol < 0) {
		# mute volume
		&volume($client, 0);
	} else {
		# un-mute volume
		&volume($client, $vol);
	}
	SliMP3::Prefs::clientSet($client, "volume", $vol);
	SliMP3::Display::volumeDisplay($client);
}

#
# tell the client to unpause the decoder
#
sub resume {
	my $client = shift;

	if (!SliMP3::Client::isSliMP3($client)) {
		return 1;
	}

	&volume($client, SliMP3::Prefs::clientGet($client, "volume"));

	if ($client->revision < 1.3) {
		streamControl($client, 'old_resume');
	} else {
		SliMP3::Stream::unpause($client);
	}

	return 1;
}

#
# pause
#
sub pause {
	my $client = shift;
	if (!SliMP3::Client::isSliMP3($client)) {
		return 1;
	}
	if ($client->revision < 1.3) {
		streamControl($client, 'old_pause');
	} else {
		SliMP3::Stream::pause($client);
	}
	return 1;
}

#
# does the same thing as pause
#
sub stop {
	my $client = shift;

	if (!SliMP3::Client::isSliMP3($client)) {
		return 1;
	}

	if ($client->revision < 1.3) {
		streamControl($client, 'old_pause');
	} else {
		SliMP3::Stream::stop($client);
	}
}

#
# playout - play out what's in the buffer
#
sub playout {
	my $client = shift;
	if (!SliMP3::Client::isSliMP3($client)) {
		return 1;
	}
	if ($client->revision < 1.3) {
		# do nothing, let the player play out, we just won't feed it anymore
	} else {
		SliMP3::Stream::playout($client);
	}
	return 1;
}

#
# send the appropriate code to the client.
#
sub streamControl {
	my ($client,$cmd) = @_;
	die unless ($client->revision <= 1.3);

	$::d_control && msg("streamControl($cmd)\n");

	my $code = $streamControlCode{$cmd};

	if ($::d_usage) {
		$code |= $streamControlCode{'old_showusage'};
	}
	$client->htmlstatusvalid(0);

	$client->waitforstart(1) if ($cmd eq 'old_start');
	$client->waitforstart(1) if ($cmd eq 'old_startpaused');
	my $success = SliMP3::Protocol::sendClient($client, 's'.$code);
	
	$::d_control && !$success && msg("problem sending $cmd to client\n");

	return $success;
}

1;

__END__
